# Stubs for pysolar.solar (Python 3.6)

import datetime
from typing import List

def solar_test() -> None: ...
def equation_of_time(day:int) -> float: ...
def get_aberration_correction(sun_earth_distance:float) -> float: ...
def get_altitude(latitude_deg:float, longitude_deg:float, when:datetime.datetime, elevation:float = ..., temperature:float = ..., pressure:float = ...) -> float: ...
def get_altitude_fast(latitude_deg:float, longitude_deg:float, when:datetime.datetime) -> float: ...
def get_apparent_sidereal_time(jd:float, jme:float, nutation_float) -> float: ...
def get_apparent_sun_longitude(geocentric_longitude:float, nutation:float, ab_correction:float) -> float: ...
def get_azimuth(latitude_deg:float, longitude_deg:float, when:datetime.datetime, elevation:float = ...) -> float: ...
def get_azimuth_fast(latitude_deg:float, longitude_deg:float, when:datetime.datetime) -> float: ...
def get_coeff(jme:float, coeffs:List[List[float]]) -> float: ...
def get_declination(day:int) -> float: ...
def get_equatorial_horizontal_parallax(sun_earth_distance:float) -> float: ...
def get_flattened_latitude(latitude:float) -> float: ...
def get_geocentric_latitude(jme:float) -> float: ...
def get_geocentric_longitude(jme:float) -> float: ...
def get_geocentric_sun_declination(apparent_sun_longitude:float, true_ecliptic_obliquity:float, geocentric_latitude:float) -> float: ...
def get_geocentric_sun_right_ascension(apparent_sun_longitude:float, true_ecliptic_obliquity:float, geocentric_latitude:float) -> float: ...
def get_heliocentric_latitude(jme:float)  -> float: ...
def get_heliocentric_longitude(jme:float) -> float: ...
def get_hour_angle(when:datetime.datetime, longitude_deg:float) -> float: ...
def get_incidence_angle(topocentric_zenith_angle:float, slope:float, slope_orientation:float, topocentric_azimuth_angle:float) -> float: ...
def get_local_hour_angle(apparent_sidereal_time:float, longitude:float, geocentric_sun_right_ascension:float) -> float: ...
def get_mean_sidereal_time(jd:float) -> float: ...
def get_nutation(jce:float) -> float: ...
def get_parallax_sun_right_ascension(projected_radial_distance:float, equatorial_horizontal_parallax:float, local_hour_angle:float, geocentric_sun_declination:float) -> float: ...
def get_projected_radial_distance(elevation:float, latitude:float) -> float: ...
def get_projected_axial_distance(elevation:float, latitude:float) -> float: ...
def get_sun_earth_distance(jme:float) -> float: ...
def get_refraction_correction(pressure:float, temperature:float, topocentric_elevation_angle:float) -> float: ...
def get_solar_time(longitude_deg:float, when:datetime.datetime) -> float: ...
def get_topocentric_azimuth_angle(topocentric_local_hour_angle:float, latitude:float, topocentric_sun_declination:float) -> float: ...
def get_topocentric_elevation_angle(latitude:float, topocentric_sun_declination:float, topocentric_local_hour_angle:float) -> float: ...
def get_topocentric_local_hour_angle(local_hour_angle:float, parallax_sun_right_ascension:float) -> float: ...
def get_topocentric_sun_declination(geocentric_sun_declination:float, projected_axial_distance:float, equatorial_horizontal_parallax:float, parallax_sun_right_ascension:float, local_hour_angle:float) -> float: ...
def get_topocentric_sun_right_ascension(projected_radial_distance:float, equatorial_horizontal_parallax:float, local_hour_angle:float, apparent_sun_longitude:float, true_ecliptic_obliquity:float, geocentric_latitude:float) -> float: ...
def get_topocentric_zenith_angle(latitude:float, topocentric_sun_declination:float, topocentric_local_hour_angle:float, pressure:float, temperature:float) -> float: ...
def get_true_ecliptic_obliquity(jme:float, nutation:float) -> float: ...



#!/usr/bin/python3

# Copyright Brandon Stafford
#
# This file is part of Pysolar.
#
# Pysolar is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Pysolar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with Pysolar. If not, see <http://www.gnu.org/licenses/>.

"""Tool for requesting data from US Naval Observatory

"""
import datetime
import random
from time import strptime

try:
  from urllib.request import Request, urlopen
  from urllib.parse import urlencode
except:
  from urllib2 import Request, urlopen
  from urllib import urlencode

from pysolar.solar import *



class Ephemeris:
    def __init__(self, timestamp, latitude, longitude, elevation, azimuth=0, altitude=0):
        self.timestamp = timestamp
        self.latitude = latitude
        self.longitude = longitude
        self.elevation = float(elevation)
        self.azimuth = float(azimuth)
        self.altitude = float(altitude)

class EphemerisComparison:
    def __init__(self, name1, eph1, name2, eph2):
        self.timestamp = eph1.timestamp
        self.latitude = eph1.latitude
        self.longitude = eph1.longitude
        self.elevation = eph1.elevation
        self.name1 = name1
        self.alt1 = eph1.altitude
        self.az1 = eph1.azimuth
        self.name2 = name2
        self.alt2 = eph2.altitude
        self.az2 = eph2.azimuth
        self.alt_error = abs(eph1.altitude - eph2.altitude)
        self.az_error = abs(eph1.azimuth - eph2.azimuth)

def RequestEphemerisData(datum):
    data = EncodeRequest(datum.latitude, datum.longitude, datum.timestamp, datum.elevation)
    url = 'http://aa.usno.navy.mil/cgi-bin/aa_topocentric2.pl'
    if type(data) == str:
      req = Request(url, data.encode())
    else:
      req = Request(url, data)
    response = urlopen(req)

    lines = response.readlines()
    response.close()
    #print(lines)
    #print(lines[21]) # should not we do some try catch here?
    result = lines[21]
    tokens = [x for x in result.split(b' ') if x not in b' ']
    print('Tokens: \n', tokens)

    usno_alt = float(tokens[4]) + float(tokens[5])/60.0 + float(tokens[6])/3600.0
    usno_az = float(tokens[7]) + float(tokens[8])/60.0 + float(tokens[9])/3600.0

#   print(usno_alt)
#   print(usno_az)

    result  = Ephemeris(datum.timestamp, datum.latitude, datum.longitude, datum.elevation, usno_az, usno_alt)

    return result

def ComparePysolarToUSNO(datum):
    alt = get_altitude(float(datum.latitude), float(datum.longitude), datum.timestamp, datum.elevation)
    pysolar_alt = (90.0 - alt)
    az = get_azimuth(float(datum.latitude), float(datum.longitude), datum.timestamp, datum.elevation)
    pysolar_az = (180.0 - az)%360.0

#   print(pysolar_alt)
#   print(pysolar_az)

    py_solar = Ephemeris(datum.timestamp, datum.latitude, datum.longitude, datum.elevation, pysolar_az, pysolar_alt)
    c = EphemerisComparison('pysolar', py_solar, 'usno', datum)
    return c

def EncodeRequest(latitude, longitude, timestamp, elevation):
    """Builds a string of arguments to be passed to the Perl script at the USNO

    Note that the degree arguments must be integers, or the USNO script chokes."""
    params = {}
    params['FFX'] = '2' # use worldwide locations script
    params['ID'] = 'pysolar'
    params['pos'] = '9'
    params['obj'] = '10' # Sun
    params['xxy'] = str(timestamp.year)
    params['xxm'] = str(timestamp.month)
    params['xxd'] = str(timestamp.day)
    params['t1'] = str(timestamp.hour)
    params['t2'] = str(timestamp.minute)
    params['t3'] = str(timestamp.second)
    params['intd'] = '1.0'
    params['unit'] = '1'
    params['rep'] = '1'
    params['place'] = 'Name omitted'

    sign = lambda x: ('1', '-1')[x < 0]
    (deg, rem) = divmod(longitude, 1)
    (min, sec) = divmod(rem, 1.0/60.0)
    params['xx0'] = sign(deg)# longitude (1 = east, -1 = west)
    params['xx1'] = str(abs(int(deg))) # degrees
    params['xx2'] = str(int(min)) # minutes
    params['xx3'] = str(sec) # seconds

    (deg, rem) = divmod(latitude, 1)
    (min, sec) = divmod(rem, 1.0/60.0)
    params['yy0'] = sign(deg) # latitude (1 = north, -1 = south)
    params['yy1'] = str(abs(int(deg))) # degrees
    params['yy2'] = str(int(min)) # minutes
    params['yy3'] = str(sec) # seconds

    params['hh1'] = str(elevation) # height above sea level in meters
    params['ZZZ'] = 'END'
    data = urlencode(params)
    return data

def GatherRandomEphemeris():
    latitude = random.randrange(-90, 90)
    longitude = random.randrange(0, 360)
    elevation = 0.0
    t = datetime.datetime(random.randrange(2013,2014), random.randrange(1, 13), random.randrange(1, 28), random.randrange(0, 24), random.randrange(0, 60), random.randrange(0,60))
    query = Ephemeris(t, latitude, longitude, elevation)
    PrintEphemerisDatum(query)
    d = RequestEphemerisData(query)
    PrintEphemerisDatum(d)
    WriteEphemerisDatumToFile(d, 'usno_data.txt')

def PrintEphemerisDatum(datum):
    print(datum.timestamp, datum.latitude, datum.longitude, datum.elevation, datum.azimuth, datum.altitude)

def ReadEphemeridesLog(logname):
    data = []
    log = open(logname, 'r')
    lines = log.readlines()
    log.close()
    for line in lines:
        args = line.split(' ')
        unaware = datetime.datetime(*(strptime(args[0] + ' ' + args[1], '%Y-%m-%d %H:%M:%S')[0:6]))
        d = unaware.replace(tzinfo=datetime.timezone.utc)
        e = Ephemeris(d, args[2], args[3], args[4], args[5], args[6])
        data.append(e)
    return data

def WriteEphemerisDatumToFile(d, filename):
    log = open(filename, 'a')
    log.write('%s %s %s %s %s %s\n' % (d.timestamp, d.latitude, d.longitude, d.elevation, d.azimuth, d.altitude))
    log.close()

def WriteComparisonsToCSV(comps, filename):
    out = open(filename, 'w')
    out.write('timestamp,latitude,longitude,elevation,alt1,alt2,alt_error,az1,az2,az_error\n')
    for c in comps:
        out.write('%s,%s,%s,%s,%s,%s,%s,%s,%s,%s\n' % (c.timestamp, c.latitude, c.longitude, c.elevation, c.alt1, c.alt2, c.alt_error, c.az1, c.az2, c.az_error))
    out.close()


if __name__ == '__main__':
    from scipy import stats
    import numpy as np
    import sys

    if len(sys.argv) >= 2:
                ephemerides = ReadEphemeridesLog(sys.argv[1])
    else:
        for i in range(100):
            GatherRandomEphemeris()
            ephemerides = ReadEphemeridesLog('usno_data.txt')

    comps = []
    for e in ephemerides:
        c = ComparePysolarToUSNO(e)
        comps.append(c)

    az_errors = np.array([c.az_error for c in comps])
    alt_errors = np.array([c.alt_error for c in comps])

    print('---------------------')
    print('Azimuth stats')
    print('Mean error: ' + str(np.mean(az_errors)))
    print('Std dev: ' + str(np.std(az_errors)))
    print('Min error: ' + str(stats.tmin(az_errors, None)))
    print('Max error: ' + str(stats.tmax(az_errors, None)))

    print('----------------------')
    print('Altitude stats')

    print('Mean error: ' + str(np.mean(alt_errors)))
    print('Std dev: '+ str(np.std(alt_errors)))
    print('Min error: ' + str(stats.tmin(alt_errors, None)))
    print('Max error: ' + str(stats.tmax(alt_errors, None)))

    WriteComparisonsToCSV(comps, 'pysolar_v_usno.csv')



# Stubs for pysolar.simulate (Python 3.6)

import datetime
from typing import Iterator, List, Tuple

def datetime_range(start_datetime:datetime.datetime, end_datetime:datetime.datetime, step_minutes:float) -> Iterator[datetime.datetime]: ...
def simulate_span(latitude_deg:float, longitude_deg:float, horizon:List[float], start_datetime:datetime.datetime, end_datetime:datetime.datetime, step_minutes:float, elevation:float = ..., temperature:float = ..., pressure:float = ...) -> Iterator[Tuple[datetime.datetime, float, float, float, float]]: ...  # TODO unclear what horizon is, and never used in the code



from functools import wraps
import inspect



class NoTimeZoneInfoError(ValueError):
  def __init__(self, argname, dt, *args):
    self.argname = argname
    self.dt = dt
    super().__init__(*args)

  def __str__(self):
    return """datetime value '{dt}' given for arg '{argname}' \
should be made timezone-aware.
You have to specify the 'tzinfo' attribute of \
'datetime.datetime' objects.""".format(dt=self.dt,
                                       argname=self.argname)


def check_aware_dt(*argnames):
  """Returns a decorator that makes sure that
  all the arguments in 'argnames', are 'datetime.datetime' objects
  with a non-null tzinfo attribute

  Parameters
  ----------
  argnames : List[str]
    list of names from the arguments of function 'func'
    (the decorated function) that are supposed to only accept
    as values 'datetime.datetime' objects
    with a non-null tzinfo attribute

  Returns
  -------
  checker : function
    decorator that makes sure that every argument of the decorated
    function that is in 'argnames' is given only 'datetime.datetime' objects
    with non-null tzinfo attribute as values"""
  def checker(func):
    """Decorator (see above in the `Returns` section)

    Parameters
    ----------
    func : function
      decorated function
      every argument of 'func' that is in argnames
      will only accept tz-aware 'datetime' objects

    Returns
    -------
    func_with_check : function
      decorated function, basically the same function as 'func',
      with the check for tz-awareness performed before execution"""
    @wraps(func)
    def func_with_check(*args, **kwargs):
      """Decorated function ; will be, apart from the check,
      completely identical to the 'func' function
      We search for 'argnames' values in args and kwargs

      Parameters
      ----------
      args : list
        list of positional arguments that are supposedly
        passed to the 'func' function
      kwargs : dict
        dict of keyword arguments that are supposedly passed
        to the 'func' function

      Returns
      -------
      func(*args, **kwargs)
        we return the very same result that would have been returned
        by the 'func' function alone
        we just checked the values of args from argnames before"""
      for argname in argnames:
        # first checking if argname is a valid arg name
        full = inspect.getfullargspec(func)
        if (argname in full.args or
            argname in full.kwonlyargs):
          # getting value given for argname
          try:
            dt = args[full.args.index(argname)]
          except (IndexError, ValueError):
            dt = kwargs[argname]
          # checking if dt is timezone-aware
          if not hasattr(dt, 'shape'):   # don't raise Exception if dt is an array (assumed of datetime64)
            if not hasattr(dt, 'tzinfo'):
              raise ValueError(
                "Expected a 'datetime.datetime' object \
    for arg '%s', got %s instead" % (argname, dt))
            if dt.tzinfo is None:
              raise NoTimeZoneInfoError(argname, dt)
      return func(*args, **kwargs)
    return func_with_check
  return checker



#    Copyright Brandon Stafford
#
#    This file is part of Pysolar.
#
#    Pysolar is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Pysolar is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with Pysolar. If not, see <http://www.gnu.org/licenses/>.

"""Support functions for horizon calculation

"""
from . import numeric as math
import datetime
from . import constants
from . import radiation
from . import solar

def datetime_range(start_datetime, end_datetime, step_minutes):
    '''yields a sequence of datetimes evenly spaced apart by step_minutes.'''
    step = step_minutes * 60
    span = end_datetime - start_datetime
    dt = datetime.timedelta(seconds = step)
    for n in range((span.days * constants.seconds_per_day + span.seconds) // step) :
        yield start_datetime + dt * n
    #end for
#end datetime_range

def simulate_span(latitude_deg, longitude_deg, horizon, start_datetime, end_datetime, step_minutes, elevation = 0, temperature = constants.standard_temperature, pressure = constants.standard_pressure):
    '''simulates the motion of the sun over a time span and location of your choosing.
    
    temperature in Kelvin and pressure in Pascal

    The start and end points are set by datetime objects, which can be created with
    the standard Python datetime module like this:
    import datetime
    start = datetime.datetime(2008, 12, 23, 23, 14, 0)
    '''
    alt_zero = 380
    for time in datetime_range(start_datetime, end_datetime, step_minutes) :
        alt = solar.get_altitude(latitude_deg, longitude_deg, time, elevation, temperature, pressure)
        azi = solar.get_azimuth(latitude_deg, longitude_deg, time, elevation)
        shade = horizon[round(azi)]
        if shade < alt_zero - round(alt_zero * math.sin(math.radians(alt))) :
            rad = 0
        else :
            rad = radiation.get_radiation_direct(time, alt)
        #end if
        yield time, alt, azi, rad, shade
    #end for
#end simulate_span

#       xs = shade.GetXShade(width, 120, azimuth_deg)
#       ys = shade.GetYShade(height, 120, altitude_deg)
#       shaded_area = xs * ys
#       shaded_percentage = shaded_area/area
# import simulate, datetime; s = datetime.datetime(2008,1,1); e = datetime.datetime(2008,1,5); simulate.simulate_span(42.0, -70.0, s, e, 30)



# -*- coding: utf-8 -*-
#    Copyright Brandon Stafford
#
#    This file is part of Pysolar.
#
#    Pysolar is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Pysolar is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with Pysolar. If not, see <http://www.gnu.org/licenses/>.

"""Additional support functions for solar geometry, astronomy, radiation correlation

:Original author: Simeon Nwaogaidu
:Contact: SimeonObinna.Nwaogaidu AT lahmeyer DOT de

:Additional author: Holger Zebner
:Contact: holger.zebner AT lahmeyer DOT de

:Additional author: Brandon Stafford

"""
from datetime import \
    datetime, \
    timedelta
from . import numeric as math
from . import solar, constants
from .tzinfo_check import check_aware_dt

# Some default constants

AM_default = 2.0             # Default air mass is 2.0
TL_default = 1.0             # Default Linke turbidity factor is 1.0
SC_default = 1367.0          # Solar constant in W/m^2 is 1367.0. Note that this value could vary by +/-4 W/m^2
TY_default = 365             # Total year number from 1 to 365 days
elevation_default = 0.0      # Default elevation is 0.0

# Useful equations for analysis

@check_aware_dt('when')
def get_sunrise_sunset_transit(latitude_deg, longitude_deg, when):
    """This function calculates the astronomical sunrise, sunset and sun transit times in local time.

    Parameters
    ----------
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting
        the north/south angular location of a place on a sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location
        in an east-west direction,relative to the Greenwich meridian.
    when : datetime.datetime
        date and time in any valid timezone, answers will be for same day in same timezone.

    Returns
    -------
    sunrise_time_dt : datetime.datetime
        Sunrise time in local time.
    sunset_time_dt : datetime.datetime
        Sunset time in local time.
    transit_time_dt:
        Sun transit time in local time.

    References
    ----------
    .. [1] http://www.skypowerinternational.com/uploads/documents/7.1415.01.121_cm121_bed-anleitung_engl.pdf
    .. [2] http://pysolar.org/
    .. [3] https://www.esrl.noaa.gov/gmd/grad/solcalc/solareqns.PDF
    .. [4] https://www.sciencedirect.com/science/article/pii/S0038092X11004592
    .. [5] https://iris.unipa.it/retrieve/handle/10447/55143/28818/Articolo.pdf

    Examples
    --------
    >>> lat = 50.111512
    >>> lon = 8.680506
    >>> timezone_local = dateutil.tz.gettz('Europe/Berlin')
    >>> now = datetime.now(timezone_local)
    >>> sr, ss, tr = sb.get_sunrise_sunset_transit(lat, lon, now)
    >>> print('sunrise: ', sr)
    >>> print('sunset:', ss)
    >>> print('transit:', tr)

    """

    utc_offset = when.utcoffset()
    if utc_offset != None :
        utc_offset = utc_offset.total_seconds()
    else :
        utc_offset = 0
    #end if

    # hours (ignore DST) time diff between UTC and the standard meridian for the input longitude 
    time_diff = int(round(longitude_deg / 15., 0) - utc_offset / 3600)
    # the day, and therefore sunrise and sunset, may be different at the input longitude than at the input "when" 
    local_time = when - timedelta(seconds = utc_offset) + timedelta(hours = time_diff)

    day = local_time.timetuple().tm_yday # Day of the year

    SHA = utc_offset / 3600 * 15.0 - longitude_deg # Solar hour angle
    TT = 2 * math.pi * day / 366
    decl = \
        (
            0.322003
        -
            22.971 * math.cos(TT)
        -
            0.357898 * math.cos(2 * TT)
        -
            0.14398 * math.cos(3 * TT)
        +
            3.94638 * math.sin(TT)
        +
            0.019334 * math.sin(2 * TT)
        +
            0.05928 * math.sin(3 * TT)
        ) # solar declination in degrees
    TT = math.radians(279.134 + 0.985647 * day) # Time adjustment angle
    time_adst = \
        (
                (
                    5.0323
                -
                    100.976 * math.sin(TT)
                +
                    595.275 * math.sin(2 * TT)
                +
                    3.6858 * math.sin(3 * TT)
                -
                    12.47 * math.sin(4 * TT)
                -
                    430.847 * math.cos(TT)
                +
                    12.5024 * math.cos(2 * TT)
                +
                    18.25 * math.cos(3 * TT)
                )
            /
                3600
        ) # Time adjustment in hours
    TON = 12 + SHA / 15.0 - time_adst # Time of noon
    ha = \
        (
            math.acos(
                math.cos(math.radians(90.833))
            /
                (
                    math.cos(math.radians(latitude_deg))
                *
                    math.cos(math.radians(decl))
                )
            -
                math.tan(math.radians(latitude_deg))
            *
                math.tan(math.radians(decl))
            )
        *
            (12 / math.pi)
        )

    # tzinfo doesn't matter
    local_day = datetime(year = local_time.year, month = local_time.month, day = local_time.day, tzinfo = local_time.tzinfo)

    transit_time = local_day + timedelta(hours = time_diff + TON)
    transit_time = (transit_time + timedelta(seconds = utc_offset) - timedelta(hours = time_diff)).replace(tzinfo=when.tzinfo)
    sunrise_time = transit_time - timedelta(hours = ha)
    sunset_time = transit_time + timedelta(hours = ha)

    return sunrise_time, sunset_time, transit_time

@check_aware_dt('when')
def get_sunrise_sunset(latitude_deg, longitude_deg, when):
    "Wrapper for get_sunrise_sunset_transit that returns just the sunrise and the sunset time."
    return \
        get_sunrise_sunset_transit(latitude_deg, longitude_deg, when)[0:2]

@check_aware_dt('when')
def get_sunrise_time(latitude_deg, longitude_deg, when):
    "Wrapper for get_sunrise_sunset_transit that returns just the sunrise time."
    return \
        get_sunrise_sunset_transit(latitude_deg, longitude_deg, when)[0]

@check_aware_dt('when')
def get_sunset_time(latitude_deg, longitude_deg, when):
    "Wrapper for get_sunrise_sunset_transit that returns just the sunset time."
    return \
        get_sunrise_sunset_transit(latitude_deg, longitude_deg, when)[1]

@check_aware_dt('when')
def get_transit_time(latitude_deg, longitude_deg, when):
    "Wrapper for get_sunrise_sunset_transit that returns just the transit time."
    return \
        get_sunrise_sunset_transit(latitude_deg, longitude_deg, when)[2]

@check_aware_dt('when')
def mean_earth_sun_distance(when):
    """Mean Earth-Sun distance is the arithmetical mean of the maximum and minimum distances
    between a planet (Earth) and the object about which it revolves (Sun). However,
    the function is used to  calculate the Mean earth sun distance.

    Parameters
    ----------
    when : datetime.datetime
        date/time for which to do the calculation

    Returns
    -------
    KD : float
        Mean earth sun distance

    References
    ----------
    .. [1] http://sunbird.jrc.it/pvgis/solres/solmod3.htm#clear-sky%20radiation
    .. [2] R. aguiar and et al, "The ESRA user guidebook, vol. 2. database", models and exploitation software-Solar
            radiation models, p.113
    """

    return 1 - 0.0335 * math.sin(2 * math.pi * (when.utctimetuple().tm_yday - 94)) / 365

@check_aware_dt('when')
def extraterrestrial_irrad(latitude_deg, longitude_deg, when, SC=SC_default):
    """Equation calculates Extratrestrial radiation. Solar radiation incident outside the earth's
    atmosphere is called extraterrestrial radiation. On average the extraterrestrial irradiance
    is 1367 Watts/meter2 (W/m2). This value varies by + or - 3 percent as the earth orbits the sun.
    The earth's closest approach to the sun occurs around January 4th and it is furthest
    from the sun around July 5th.

    Parameters
    ----------
    when : datetime.datetime
        date/time for which to do the calculation
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting the north/south angular location
        of a place on a sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location in an east-west direction,relative
        to the Greenwich meridian.
    SC : float
        The solar constant is the amount of incoming solar electromagnetic radiation per unit area, measured
        on the outer surface of Earth's atmosphere in a plane perpendicular to the rays.It is measured by
        satellite to be roughly 1366 watts per square meter (W/m^2)

    Returns
    -------
    EXTR1 : float
        Extraterrestrial irradiation

    References
    ----------
    .. [1] http://solardat.uoregon.edu/SolarRadiationBasics.html
    .. [2] Dr. J. Schumacher and et al,"INSEL LE(Integrated Simulation Environment Language)Block reference",p.68

    """
    day = when.utctimetuple().tm_yday
    ab = math.cos(2 * math.pi * (day - 1.0)/(365.0))
    bc = math.sin(2 * math.pi * (day - 1.0)/(365.0))
    cd = math.cos(2 * (2 * math.pi * (day - 1.0)/(365.0)))
    df = math.sin(2 * (2 * math.pi * (day - 1.0)/(365.0)))
    decl = math.radians(solar.get_declination(day))
    ha = math.radians(solar.get_hour_angle(when, longitude_deg))
    ZA = math.sin(math.radians(latitude_deg)) * math.sin(decl) + math.cos(math.radians(latitude_deg)) * math.cos(decl) * math.cos(ha)

    return SC * ZA * (1.00010 + 0.034221 * ab + 0.001280 * bc + 0.000719 * cd + 0.000077 * df) if ZA > 0 else 0.0


@check_aware_dt('when')
def declination_degree(when, TY = TY_default):
    """The declination of the sun is the angle between Earth's equatorial plane and a line
    between the Earth and the sun. It varies between 23.45 degrees and -23.45 degrees,
    hitting zero on the equinoxes and peaking on the solstices.

    Parameters
    ----------
    when : datetime.datetime
        date/time for which to do the calculation
    TY : float
        Total number of days in a year. eg. 365 days per year,(no leap days)

    Returns
    -------
    DEC : float
        The declination of the Sun

    References
    ----------
    .. [1] http://pysolar.org/

    """
    return constants.earth_axis_inclination * math.sin((2 * math.pi / (TY)) * ((when.utctimetuple().tm_yday) - 81))


@check_aware_dt('when')
def solarelevation_function_clear(latitude_deg, longitude_deg, when, temperature = constants.standard_temperature,
                                  pressure = constants.standard_pressure,  elevation = elevation_default):
    """Equation calculates Solar elevation function for clear sky type.

    Parameters
    ----------
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting
        the north/south angular location of a place on a sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location
        in an east-west direction,relative to the Greenwich meridian.
    when : datetime.datetime
        date/time for which to do the calculation
    temperature : float
        atmospheric temperature in kelvin
   pressure : float
        pressure in pascals
    elevation : float
        The elevation of a geographic location is its height above a fixed reference point, often the mean
        sea level.

    Returns
    -------
    SOLALTC : float
        Solar elevation function clear sky

    References
    ----------
    .. [1] S. Younes, R.Claywell and el al,"Quality control of solar radiation data: present status
            and proposed new approaches", energy 30 (2005), pp 1533 - 1549.

    """
    altitude = solar.get_altitude(latitude_deg, longitude_deg,when, elevation, temperature,pressure)
    return (0.038175 + (1.5458 * (math.sin(altitude))) + ((-0.59980) * (0.5 * (1 - math.cos(2 * (altitude))))))

@check_aware_dt('when')
def solarelevation_function_overcast(latitude_deg, longitude_deg, when,
                                     elevation = elevation_default, temperature = constants.standard_temperature,
                                     pressure = constants.standard_pressure):
    """ The function calculates solar elevation function for overcast sky type.
    This associated hourly overcast radiation model is based on the estimation of the
    overcast sky transmittance with the sun directly overhead combined with the application
    of an over sky elavation function to estimate the overcast day global irradiation
    value at any solar elevation.

    Parameters
    ----------
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting the north/south angular location of a place on a
        sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location in an east-west direction,relative to the
        Greenwich meridian.
    when : datetime.datetime
        date/time for which to do the calculation
    elevation : float
        The elevation of a geographic location is its height above a fixed reference point, often the mean sea level.
    temperature : float
        atmospheric temperature in kelvin
    pressure : float
        pressure in pascals

    Returns
    -------
    SOLALTO : float
        Solar elevation function overcast

    References
    ----------
    .. [1] Prof. Peter Tregenza,"Solar radiation and daylight models", p.89.

    .. [2] Also accessible through Google Books: http://tinyurl.com/5kdbwu
        Tariq Muneer, "Solar Radiation and Daylight Models, Second Edition: For the Energy Efficient
        Design of Buildings"

    """
    altitude = solar.get_altitude(latitude_deg, longitude_deg,when, elevation, temperature,pressure)
    return ((-0.0067133) + (0.78600 * (math.sin(altitude)))) + (0.22401 * (0.5 * (1 - math.cos(2 * altitude))))


def diffuse_transmittance(TL = TL_default):
    """Equation calculates the Diffuse_transmittance and the is the Theoretical Diffuse Irradiance on a horizontal
    surface when the sun is at the zenith.

    Parameters
    ----------
    TL : float
        Linke turbidity factor

    Returns
    -------
    DT : float
        diffuse_transmittance

    References
    ----------
    .. [1] S. Younes, R.Claywell and el al,"Quality control of solar radiation data: present status and proposed
            new approaches", energy 30 (2005), pp 1533 - 1549.

    """
    return ((-21.657) + (41.752 * (TL)) + (0.51905 * (TL) * (TL)))


@check_aware_dt('when')
def diffuse_underclear(latitude_deg, longitude_deg, when, elevation = elevation_default,
                       temperature = constants.standard_temperature, pressure = constants.standard_pressure, TL=TL_default):
    """Equation calculates diffuse radiation under clear sky conditions.

    Parameters
    ----------
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting the north/south angular location of a place on
        a sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location in an east-west direction,relative to the
        Greenwich meridian.
    when : datetime.datetime
        date/time for which to do the calculation
    elevation : float
        The elevation of a geographic location is its height above a fixed reference point, often the mean sea level.
    temperature : float
        atmospheric temperature in kelvin
    pressure : float
        pressure in pascals
    TL : float
        Linke turbidity factor

    Returns
    -------
    DIFFC : float
        Diffuse Irradiation under clear sky

    References
    ----------
    .. [1] S. Younes, R.Claywell and el al,"Quality control of solar radiation data: present status and proposed
            new approaches", energy 30 (2005), pp 1533 - 1549.

    """
    DT = ((-21.657) + (41.752 * (TL)) + (0.51905 * (TL) * (TL)))
    altitude = solar.get_altitude(latitude_deg, longitude_deg,when, elevation, temperature,pressure)

    return mean_earth_sun_distance(when) * DT * altitude

@check_aware_dt('when')
def diffuse_underovercast(latitude_deg, longitude_deg, when, elevation = elevation_default,
                          temperature = constants.standard_temperature, pressure = constants.standard_pressure,TL=TL_default):
    """Function calculates the diffuse radiation under overcast conditions.

    Parameters
    ----------
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting the north/south angular location of a place on a
        sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location in an east-west direction,relative to the
        Greenwich meridian.
    when : datetime.datetime
        date/time for which to do the calculation
    elevation : float
        The elevation of a geographic location is its height above a fixed reference point, often the mean sea level.
    temperature : float
        atmospheric temperature in kelvin
    pressure : float
        pressure in pascals
    TL : float
        Linke turbidity factor

    Returns
    -------
    DIFOC : float
        Diffuse Irradiation under overcast

    References
    ----------
    .. [1] S. Younes, R.Claywell and el al,"Quality control of solar radiation data: present status and proposed
            new approaches", energy 30 (2005), pp 1533 - 1549.

    """
    DT = ((-21.657) + (41.752 * (TL)) + (0.51905 * (TL) * (TL)))

    DIFOC = ((mean_earth_sun_distance(when)
              )*(DT)*(solar.get_altitude(latitude_deg,longitude_deg, when, elevation,
                                        temperature, pressure)))
    return DIFOC

@check_aware_dt('when')
def direct_underclear(latitude_deg, longitude_deg, when,
                      TY = TY_default, AM = AM_default, TL = TL_default, elevation = elevation_default, 
                      temperature = constants.standard_temperature, pressure = constants.standard_pressure):
    """Equation calculates direct radiation under clear sky conditions.

    Parameters
    ----------
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting the north/south angular location of a
        place on a sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location in an east-west direction,relative to the
        Greenwich meridian.
    when : datetime.datetime
        date/time for which to do the calculation
    TY : float
        Total number of days in a year. eg. 365 days per year,(no leap days)
    AM : float
        Air mass. An Air Mass is a measure of how far light travels through the Earth's atmosphere. One air mass,
        or AM1, is the thickness of the Earth's atmosphere. Air mass zero (AM0) describes solar irradiance in space,
        where it is unaffected by the atmosphere. The power density of AM1 light is about 1,000 W/m^2
    TL : float
        Linke turbidity factor
    elevation : float
        The elevation of a geographic location is its height above a fixed reference point, often the mean
        sea level.
    temperature : float
        atmospheric temperature in kelvin
    pressure : float
        pressure in pascals

    Returns
    -------
    DIRC : float
        Direct Irradiation under clear

    References
    ----------
    .. [1] S. Younes, R.Claywell and el al,"Quality control of solar radiation data: present status and proposed
           new approaches", energy 30 (2005), pp 1533 - 1549.

    """
    KD = mean_earth_sun_distance(when)

    DEC = declination_degree(when,TY)

    DIRC = (1367 * KD * math.exp(-0.8662 * (AM) * (TL) * (DEC)
                             ) * math.sin(solar.get_altitude(latitude_deg,longitude_deg,
                                                          when,elevation ,
                                                          temperature , pressure )))

    return DIRC

@check_aware_dt('when')
def global_irradiance_clear(latitude_deg, longitude_deg, when,
                            TY = TY_default, AM = AM_default, TL = TL_default, elevation = elevation_default, 
                            temperature = constants.standard_temperature, pressure = constants.standard_pressure):

    """Equation calculates global irradiance under clear sky conditions.

    Parameters
    ----------
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting the north/south angular location of a place
        on a sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location in an east-west direction,relative to
        the Greenwich meridian.
    when : datetime.datetime
        date/time for which to do the calculation
    temperature : float
        atmospheric temperature in kelvin
    pressure : float
        pressure in pascals
    elevation : float
        The elevation of a geographic location is its height above a fixed reference point, often the
        mean sea level.
    TY : float
        Total number of days in a year. eg. 365 days per year,(no leap days)
    AM : float
        Air mass. An Air Mass is a measure of how far light travels through the Earth's atmosphere. One air mass,
        or AM1, is the thickness of the Earth's atmosphere. Air mass zero (AM0) describes solar irradiance in
        space, where it is unaffected by the atmosphere. The power density of AM1 light is about 1,000 W/m.

    TL : float
        Linke turbidity factor
    elevation : float
        The elevation of a geographic location is its height above a fixed reference point, often the mean sea
        level.

    Returns
    -------
    ghic : float
        Global Irradiation under clear sky

    References
    ----------
    .. [1] S. Younes, R.Claywell and el al,"Quality control of solar radiation data: present status and proposed
            new approaches", energy 30 (2005), pp 1533 - 1549.

    """
    DIRC =  direct_underclear(latitude_deg, longitude_deg, when,
                              TY, AM, TL, elevation, temperature = constants.standard_temperature,
                              pressure = constants.standard_pressure)

    DIFFC = diffuse_underclear(latitude_deg, longitude_deg, when,
                               elevation, temperature = constants.standard_temperature, 
                               pressure= constants.standard_pressure)

    ghic = (DIRC + DIFFC)

    return ghic


@check_aware_dt('when')
def global_irradiance_overcast(latitude_deg, longitude_deg, when,
                               elevation = elevation_default, temperature = constants.standard_temperature,
                               pressure = constants.standard_pressure):
    """Calculated Global is used to compare to the Diffuse under overcast conditions.
    Under overcast skies, global and diffuse are expected to be equal due to the absence of the beam
    component.

    Parameters
    ----------
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting the north/south angular location of a
        place on a sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location in an east-west direction,relative
        to the Greenwich meridian.
    when : datetime.datetime
        date/time for which to do the calculation
    elevation : float
        The elevation of a geographic location is its height above a fixed reference point, often the
        mean sea level.
    temperature : float
        atmospheric temperature in kelvin
    pressure : float
        pressure in pascals

    Returns
    -------
    ghioc : float
        Global Irradiation under overcast sky

    References
    ----------
    .. [1] S. Younes, R.Claywell and el al, "Quality
            control of solar radiation data: present status
            and proposed new approaches", energy 30
            (2005), pp 1533 - 1549.

    """
    ghioc = (572 * (solar.get_altitude(latitude_deg, longitude_deg, when,
                                    elevation , temperature , pressure )))

    return ghioc


def diffuse_ratio(DIFF_data, ghi_data):
    """Function calculates the Diffuse ratio.

    Parameters
    ----------
    DIFF_data : array_like
        Diffuse horizontal irradiation data
    ghi_data : array_like
        global horizontal irradiation data array

    Returns
    -------
    K : float
        diffuse_ratio

    References
    ----------
    .. [1] S. Younes, R.Claywell and el al,"Quality control of solar radiation data: present status and proposed
            new approaches", energy 30 (2005), pp 1533 - 1549.

    """
    K = DIFF_data/ghi_data

    return K


@check_aware_dt('when')
def clear_index(ghi_data, latitude_deg, longitude_deg, when):

    """This calculates the clear index ratio.

    Parameters
    ----------
    ghi_data : array_like
        global horizontal irradiation data array
    when : datetime.datetime
        date/time for which to do the calculation
    latitude_deg : float
        latitude in decimal degree. A geographical term denoting the north/south angular location of a place
        on a sphere.
    longitude_deg : float
        longitude in decimal degree. Longitude shows your location in an east-west direction,relative to the
        Greenwich meridian.

    Returns
    -------
    KT : float
        Clear index ratio

    References
    ----------
    .. [1] S. Younes, R.Claywell and el al,"Quality control of solar radiation data: present status and proposed
            new approaches", energy 30 (2005), pp 1533 - 1549.

    """
    EXTR1 = extraterrestrial_irrad(latitude_deg, longitude_deg, when)

    KT = (ghi_data / EXTR1)

    return KT



'''
Created on Aug 13, 2015

@author: rene
'''
import unittest
import datetime
from pysolar.solar import *
from pysolar.solartime import *


class Test(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testHourAngle(self):

        '''
        Example 1.6.1 from Book "Solar Engineering of Thermal Processes by Duffie and Beckman, fourth edition, Wiley 2013"

        hour angle should be -22.5. (-22.5 % 360.0 = 337.5)
        '''

        tz = datetime.timezone(datetime.timedelta(hours=-6))
        when = datetime.datetime(2008, 2, 13, 10, 42, tzinfo=tz)

        latitude_deg = 43.076342
        longitude_deg = -89.384448

        '''
        Hour angle from get_hour_angle
        '''
        ha = get_hour_angle(when, longitude_deg)

        '''Hour angle as in pysolar.solar.get_altitude'''

        # time-dependent calculations
        jd = get_julian_solar_day(when)
        jde = get_julian_ephemeris_day(when)
        jce = get_julian_ephemeris_century(jde)
        jme = get_julian_ephemeris_millennium(jce)
        geocentric_latitude = get_geocentric_latitude(jme)
        geocentric_longitude = get_geocentric_longitude(jme)
        sun_earth_distance = get_sun_earth_distance(jme)
        aberration_correction = get_aberration_correction(sun_earth_distance)
        nutation = get_nutation(jce)
        apparent_sidereal_time = get_apparent_sidereal_time(jd, jme, nutation)
        true_ecliptic_obliquity = get_true_ecliptic_obliquity(jme, nutation)

        # calculations dependent on location and time
        apparent_sun_longitude = get_apparent_sun_longitude(
            geocentric_longitude,
            nutation,
            aberration_correction)
        geocentric_sun_right_ascension = get_geocentric_sun_right_ascension(
            apparent_sun_longitude,
            true_ecliptic_obliquity,
            geocentric_latitude)
        local_hour_angle = get_local_hour_angle(
            apparent_sidereal_time,
            longitude_deg,
            geocentric_sun_right_ascension)

        reference_ha = -22.5 % 360.0

        self.assertAlmostEqual(local_hour_angle % 360.0, ha % 360.0, places=0)
        self.assertAlmostEqual(reference_ha % 360.0, ha % 360.0, places=0)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#    Copyright Sean T. Hammond
#
#    This file is part of Pysolar.
#
#    Pysolar is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Pysolar is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with Pysolar. If not, see <http://www.gnu.org/licenses/>.

"""Various elevation-related calculations

"""
import warnings
from .constants import \
    standard_pressure, \
    standard_temperature, \
    earth_temperature_lapse_rate, \
    air_gas_constant, \
    earth_gravity, \
    earth_atmosphere_molar_mass

def get_pressure_with_elevation(h, Ps=standard_pressure, Ts=standard_temperature, Tl=earth_temperature_lapse_rate, Hb=0.0, R=air_gas_constant, g=earth_gravity, M=earth_atmosphere_molar_mass):
    "This function returns an estimate of the pressure in pascals as a function of\n" \
    " elevation above sea level.\n" \
    "NOTES:\n" \
    "  * This equation is only accurate up to 11,000 meters\n" \
    "  * results might be odd for elevations below 0 (sea level), like Dead Sea.\n" \
    "h=elevation relative to sea level (m)\n" \
    "Ps= static pressure (pascals)\n" \
    "Ts= temperature (kelvin)\n" \
    "Tl= temperature lapse rate (kelvin/meter)\n" \
    "Hb= height at the bottom of the layer\n" \
    "R= universal gas constant for air\n" \
    "g= gravitational acceleration\n" \
    "M= Molar mass of atmosphere\n" \
    "P = Ps * (Ts / ((Ts + Tl) * (h - Hb))) ^ ((g * M)/(R * Tl))\n" \
    "returns pressure in pascals\n"
    if h > 11000.0 :
        warnings.warn \
          (
            "Elevation used exceeds the recommended maximum elevation for this function (11,000m)\n"
          )
    #end if
    return \
        Ps * (Ts / (Ts + Tl * (h - Hb))) ** ((g * M) / (R * Tl))
#end get_pressure_with_elevation

def get_temperature_with_elevation(h, Ts=standard_temperature, Tl=earth_temperature_lapse_rate):
    "This function returns an estimate of temperature as a function above sea level.\n" \
    "NOTES:\n" \
    "  * This equation is only accurate up to 11,000 meters\n" \
    "  * results might be odd for elevations below 0 (sea level), like Dead Sea.\n" \
    "Ts= temperature (kelvin)\n" \
    "Tl= temperature lapse rate (kelvin/meter)\n" \
    "returns temp in kelvin\n"
    return \
        Ts + h *Tl
#end get_temperature_with_elevation

def elevation_test():
    print("Elevation(m) Pressure(Pa) Temperature(K)")
    h = 0
    for _ in range(11):
        P = get_pressure_with_elevation(h)
        T = get_temperature_with_elevation(h)
        print("%i %i %i" % (h, P, T))
        h += 1000
    #end for
#end elevation_test




#    Copyright François Steinmetz
#
#    This file is part of Pysolar.
#
#    Pysolar is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Pysolar is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with Pysolar. If not, see <http://www.gnu.org/licenses/>.


"""
Import math functions from either numpy (in order to vectorize operations) or
builtins math module.

By default, use numpy when available.

To force builtins math module usage when numpy is available:
    import pysolar
    pysolar.use_math()
"""

from math import degrees, cos, sin, radians, tan, pi
from math import acos, atan, asin, atan2, exp, e

current_mod = 'math'


def globals_import_from(module, name, name_as):
    """
    Does "from <module> import <name> as <name_as>" (globally)
    """
    module = __import__(module, fromlist=[name])
    globals()[name_as] = getattr(module, name)


def where_math(condition, x, y):
    """ scalar version of numpy.where """
    if condition:
        return x
    else:
        return y

where = where_math


def tm_yday_math(d):
    return d.utctimetuple().tm_yday

tm_yday = tm_yday_math


def tm_yday_numpy(d):
    dd = numpy.array(d, dtype='datetime64[D]')
    dy = numpy.array(d, dtype='datetime64[Y]')
    return (dd - dy).astype('int') + 1


def tm_hour_math(d):
    return d.utctimetuple().tm_hour

tm_hour = tm_hour_math


def tm_hour_numpy(d):
    dh = numpy.array(d, dtype='datetime64[h]')
    dd = numpy.array(d, dtype='datetime64[D]')
    return (dh - dd).astype('int')


def tm_min_math(d):
    return d.utctimetuple().tm_min

tm_min = tm_min_math


def tm_min_numpy(d):
    dm = numpy.array(d, dtype='datetime64[m]')
    dh = numpy.array(d, dtype='datetime64[h]')
    return (dm - dh).astype('int')


def use_numpy():
    """
    Import required functions/constants from numpy
    """
    globals_import_from('numpy', 'degrees', 'degrees')
    globals_import_from('numpy', 'cos', 'cos')
    globals_import_from('numpy', 'sin', 'sin')
    globals_import_from('numpy', 'radians', 'radians')
    globals_import_from('numpy', 'tan', 'tan')
    globals_import_from('numpy', 'pi', 'pi')
    globals_import_from('numpy', 'arccos', 'acos')
    globals_import_from('numpy', 'arctan', 'atan')
    globals_import_from('numpy', 'arcsin', 'asin')
    globals_import_from('numpy', 'arctan2', 'atan2')
    globals_import_from('numpy', 'exp', 'exp')
    globals_import_from('numpy', 'e', 'e')
    globals_import_from('numpy', 'where', 'where')
    globals()['tm_yday'] = tm_yday_numpy
    globals()['tm_hour'] = tm_hour_numpy
    globals()['tm_min'] = tm_min_numpy
    globals()['current_mod'] = 'numpy'


def use_math():
    """
    Import required functions/constants from builtins math module
    """
    globals_import_from('math', 'degrees', 'degrees')
    globals_import_from('math', 'cos', 'cos')
    globals_import_from('math', 'sin', 'sin')
    globals_import_from('math', 'radians', 'radians')
    globals_import_from('math', 'tan', 'tan')
    globals_import_from('math', 'pi', 'pi')
    globals_import_from('math', 'acos', 'acos')
    globals_import_from('math', 'atan', 'atan')
    globals_import_from('math', 'asin', 'asin')
    globals_import_from('math', 'atan2', 'atan2')
    globals_import_from('math', 'exp', 'exp')
    globals_import_from('math', 'e', 'e')
    globals()['where'] = where
    globals()['tm_yday'] = tm_yday_math
    globals()['tm_hour'] = tm_hour_math
    globals()['tm_min'] = tm_min_math
    globals()['current_mod'] = 'math'


try:
    import numpy
    use_numpy()
except ImportError:
    pass



#    Copyright Brandon Stafford
#
#    This file is part of Pysolar.
#
#    Pysolar is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Pysolar is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with Pysolar. If not, see <http://www.gnu.org/licenses/>.

"""This test file makes sure that every function that requires
'datetime.datetime' objects as parameters
receives only timezone-aware ones.
"""

from pysolar import solar
from pysolar import solartime
from pysolar import util
import datetime
import unittest
from pysolar.tzinfo_check import NoTimeZoneInfoError
import numpy as np


class TestErrorTimeZoneIsNone(unittest.TestCase):
  unaware = datetime.datetime(2000, 1, 1)

  lat = 1.0
  lon = 1.0

  def test_solar_topocentric_position_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solar.get_topocentric_position(self.lat, self.lon, self.unaware)

  def test_solar_get_position_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solar.get_position(self.lat, self.lon, self.unaware)

  def test_solar_get_altitude_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solar.get_altitude(self.lat, self.lon, self.unaware)

  def test_solar_get_altitude_fast_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solar.get_altitude_fast(self.lat, self.lon, self.unaware)

  def test_solar_get_azimuth_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solar.get_azimuth(self.lat, self.lon, self.unaware)

  def test_solar_get_azimuth_fast_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solar.get_azimuth_fast(self.lat, self.lon, self.unaware)

  def test_solar_get_hour_angle_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solar.get_hour_angle(self.unaware, self.lon)

  def test_solar_get_solar_time_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solar.get_hour_angle(self.unaware, self.lon)

  def test_solartime_get_leap_seconds_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solartime.get_leap_seconds(self.unaware)

  def test_solartime_get_delta_t_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solartime.get_delta_t(self.unaware)

  def test_solartime_get_julian_solar_day_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solartime.get_julian_solar_day(self.unaware)

  def test_solartime_get_julian_ephemeris_day_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      solartime.get_julian_ephemeris_day(self.unaware)

  def test_util_get_sunrise_sunset_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.get_sunrise_sunset(self.lat, self.lon, self.unaware)

  def test_util_get_sunrise_time_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.get_sunrise_time(self.lat, self.lon, self.unaware)

  def test_util_get_sunset_time_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.get_sunset_time(self.lat, self.lon, self.unaware)

  def test_util_get_transit_time_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.get_transit_time(self.lat, self.lon, self.unaware)

  def test_util_mean_earth_sun_distance_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.mean_earth_sun_distance(self.unaware)

  def test_util_extraterrestrial_irrad_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.extraterrestrial_irrad(self.lat, self.lon, self.unaware)

  def test_util_declination_degree_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.declination_degree(self.unaware)

  def test_util_solarelevation_function_overcast_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.solarelevation_function_overcast(self.lat, self.lon, self.unaware)

  def test_util_diffuse_underclear_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.diffuse_underclear(self.lat, self.lon, self.unaware)

  def test_util_diffuse_underovercast_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.diffuse_underovercast(self.lat, self.lon, self.unaware)

  def test_util_direct_underclear_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.direct_underclear(self.lat, self.lon, self.unaware)

  def test_util_global_irradiance_clear_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.global_irradiance_clear(self.lat, self.lon, self.unaware)

  def test_util_global_irradiance_overcast_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      util.global_irradiance_overcast(self.lat, self.lon, self.unaware)

  def test_util_clear_index_raise_error(self):
    with self.assertRaises(NoTimeZoneInfoError):
      ghi_data = np.asarray([0, 0]) # Don't know what ghi_data is supposed to be
      util.clear_index(ghi_data, self.lat, self.lon, self.unaware)



class TestTimeZoneNotNone(unittest.TestCase):
  aware = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)

  lat = 1.0
  lon = 1.0

  def test_solar_topocentric_position_no_error(self):
    try:
      solar.get_topocentric_position(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solar_get_position_no_error(self):
    try:
      solar.get_position(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solar_get_altitude_no_error(self):
    try:
      solar.get_altitude(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solar_get_altitude_fast_no_error(self):
    try:
      solar.get_altitude_fast(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solar_get_azimuth_no_error(self):
    try:
      solar.get_azimuth(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solar_get_azimuth_fast_no_error(self):
    try:
      solar.get_azimuth_fast(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solar_get_hour_angle_no_error(self):
    try:
      solar.get_hour_angle(self.aware, self.lon)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solar_get_solar_time_no_error(self):
    try:
      solar.get_hour_angle(self.aware, self.lon)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solartime_get_leap_seconds_no_error(self):
    try:
      solartime.get_leap_seconds(self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solartime_get_delta_t_no_error(self):
    try:
      solartime.get_delta_t(self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solartime_get_julian_solar_day_no_error(self):
    try:
      solartime.get_julian_solar_day(self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_solartime_get_julian_ephemeris_day_no_error(self):
    try:
      solartime.get_julian_ephemeris_day(self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_get_sunrise_sunset_no_error(self):
    try:
      util.get_sunrise_sunset(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_get_sunrise_time_no_error(self):
    try:
      util.get_sunrise_time(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_get_sunset_time_no_error(self):
    try:
      util.get_sunset_time(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_mean_earth_sun_distance_no_error(self):
    try:
      util.mean_earth_sun_distance(self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_extraterrestrial_irrad_no_error(self):
    try:
      util.extraterrestrial_irrad(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_declination_degree_no_error(self):
    try:
      util.declination_degree(self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_solarelevation_function_overcast_no_error(self):
    try:
      util.solarelevation_function_overcast(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_diffuse_underclear_no_error(self):
    try:
      util.diffuse_underclear(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_diffuse_underovercast_no_error(self):
    try:
      util.diffuse_underovercast(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_direct_underclear_no_error(self):
    try:
      util.direct_underclear(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_global_irradiance_clear_no_error(self):
    try:
      util.global_irradiance_clear(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_global_irradiance_overcast_no_error(self):
    try:
      util.global_irradiance_overcast(self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")

  def test_util_clear_index(self):
    try:
      ghi_data = np.asarray([0, 0]) # Don't know what ghi_data is supposed to be
      util.clear_index(ghi_data, self.lat, self.lon, self.aware)
    except NoTimeZoneInfoError:
      self.fail("""'NoTimeZoneInfoError' should not be raised \
as 'datetime' object is tz-aware.""")



# Stubs for pysolar.time (Python 3.6)

import datetime
from typing import List, Tuple

julian_day_offset: float
gregorian_day_offset: int
tt_offset: float
leap_seconds_base_year: int
leap_seconds_adjustments: List[Tuple[float,float]]

def get_leap_seconds(when:datetime.datetime) -> int: ...

delta_t_base_year: int
delta_t_base_month: int
delta_t: List[List[float]]

def get_delta_t(when:datetime.datetime) -> float: ...
def get_julian_solar_day(when:datetime.datetime) -> float: ...
def get_julian_ephemeris_day(when:datetime.datetime) -> float: ...
def get_julian_century(julian_day:float) -> float: ...
def get_julian_ephemeris_century(julian_ephemeris_day:float) -> float: ...
def get_julian_ephemeris_millennium(julian_ephemeris_century) -> float: ...



#    Copyright Brandon Stafford
#
#    This file is part of Pysolar.
#
#    Pysolar is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Pysolar is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with Pysolar. If not, see <http://www.gnu.org/licenses/>.

# Disable linting as this is a work in progress
# flake8: noqa

from . import numeric as math
from .constants import standard_pressure

# single-scattering albedo used to calculate aerosol scattering transmittance;
# not used for sky or ground albedo for backscattering estimate
albedo = {}

albedo["high-frequency"] = 0.92
albedo["low-frequency"] = 0.84
standard_pressure_millibars = standard_pressure / 100
un = 0.0003  # atm*cm, from [Gueymard 2008], p. 280

E0n = {"high-frequency": 635.4,  # extra-atmospheric irradiance, 290-700 nm (UV and visible)
       "low-frequency":  709.7}  # extra-atmospheric irradiance, 700-4000 nm (short infrared)


def get_aerosol_forward_scatterance_factor(altitude_deg):
    Z = 90 - altitude_deg
    return 1 - math.e ** (-0.6931 - 1.8326 * math.cos(math.radians(Z)))


def get_aerosol_optical_depth(turbidity_beta, effective_wavelength, turbidity_alpha):
    # returns tau_a
    return turbidity_beta * effective_wavelength ** -turbidity_alpha

def get_aerosol_scattering_correction_factor(band, ma, tau_a):
    # returns F
    if band == "high-frequency":
        g0 = (3.715 + 0.368 * ma + 0.036294 * ma ** 2) / \
            (1 + 0.0009391 * ma ** 2)
        g1 = (-0.164 - 0.72567 * ma + 0.20701 * ma ** 2) / \
            (1 + 0.001901 * ma ** 2)
        g2 = (-0.052288 + 0.31902 * ma + 0.17871 * ma ** 2) / \
            (1 + 0.0069592 * ma ** 2)
        return (g0 + g1 * tau_a) / (1 + g2 * tau_a)
    else:
        h0 = (3.4352 + 0.65267 * ma + 0.00034328 * ma ** 2) / \
            (1 + 0.034388 * ma ** 1.5)
        h1 = (1.231 - 1.63853 * ma + 0.20667 * ma ** 2) / \
            (1 + 0.1451 * ma ** 1.5)
        h2 = (0.8889 - 0.55063 * ma + 0.50152 * ma ** 2) / \
            (1 + 0.14865 * ma ** 1.5)
        return (h0 + h1 * tau_a) / (1 + h2 * tau_a)


def get_aerosol_transmittance(band, ma, tau_a):
    # returns Ta
    return math.exp(-ma * tau_a)


def get_aerosol_scattering_transmittance(band, ma, tau_a):
    # returns Tas
    return math.exp(-ma * albedo[band] * tau_a)

def get_backscattered_diffuse_broadband_irradiance(band, turbidity_alpha=1.3, turbidity_beta=0.6):
    return get_backscattered_diffuse_broadband_irradiance_by_band("high-frequency", turbidity_alpha, turbidity_beta) + get_backscattered_diffuse_broadband_irradiance_by_band("low-frequency", turbidity_alpha, turbidity_beta)

def get_backscattered_diffuse_irradiance_by_band(band, Ebi, Edpi, turbidity_alpha=1.3, turbidity_beta=0.6):
    rhos = get_sky_albedo(band, turbidity_alpha, turbidity_beta)
    rhog = get_ground_albedo(band)
    Eddi = rhog * rhos * (Ebi + Edpi) / (1 - rhog * rhos)
    return Eddi


def get_beam_broadband_irradiance(altitude_deg, pressure_millibars=standard_pressure_millibars, ozone_atm_cm=0.35, nitrogen_atm_cm=0.0002, precipitable_water_cm=5.0, turbidity_alpha=1.3, turbidity_beta=0.6):
    Z = 90 - altitude_deg
    Ebn = get_broadband_direct_normal_irradiance(altitude_deg, pressure_millibars, ozone_atm_cm, nitrogen_atm_cm, precipitable_water_cm, turbidity_alpha, turbidity_beta)
    return Ebn * math.cos(math.radians(Z))

def get_beam_irradiance_by_band(band, altitude_deg, pressure_millibars=standard_pressure_millibars, ozone_atm_cm=0.35, nitrogen_atm_cm=0.0002, precipitable_water_cm=5.0, turbidity_alpha=1.3, turbidity_beta=0.6):
    Z = 90 - altitude_deg
    Ebni = get_direct_normal_irradiance_by_band(band, altitude_deg, pressure_millibars, ozone_atm_cm, nitrogen_atm_cm, precipitable_water_cm, turbidity_alpha, turbidity_beta)
    return Ebni * math.cos(math.radians(Z))

def get_diffuse_broadband_irradiance(air_mass=1.66, turbidity_alpha=1.3, turbidity_beta=0.6):
    return get_diffuse_irradiance_by_band("high-frequency", air_mass, turbidity_alpha, turbidity_beta) + get_diffuse_irradiance_by_band("low-frequency", air_mass, turbidity_alpha, turbidity_beta)


def get_diffuse_irradiance_by_band(band, air_mass=1.66, turbidity_alpha=1.3, turbidity_beta=0.6):
    Z = 90 - altitude_deg
    effective_wavelength = get_effective_aerosol_wavelength(band, turbidity_alpha)
    tau_a = get_aerosol_optical_depth(
        turbidity_beta, effective_wavelength, turbidity_alpha)
    ma = get_optical_mass_aerosol(altitude_deg)
    mo = get_optical_mass_ozone(altitude_deg)
    mR = get_optical_mass_rayleigh(altitude_deg, pressure_millibars)

    To = get_ozone_transmittance(band, mo)
    Tg = get_gas_transmittance(band, mR)
    Tn = get_nitrogen_transmittance(band, 1.66)
    Tw = get_water_vapor_transmittance(band, 1.66)
    TR = get_rayleigh_transmittance(band, mR)
    Ta = get_aerosol_transmittance(band, ma, tau_a)
    Tas = get_aerosol_scattering_transmittance(band, ma, tau_a)

    BR = get_rayleigh_extinction_forward_scattering_fraction(band, air_mass)
    Ba = get_aerosol_forward_scatterance_factor(altitude_deg)
    F = get_aerosol_scattering_correction_factor(band, ma, tau_a)

    Edpi = To * Tg * Tn * Tw * \
        (BR * (1 - TR) * Ta ** 0.25 + Ba * F * TR * (1 - Tas ** 0.25)) * \
        E0n[band]
    return Edpi


def get_broadband_direct_normal_irradiance(altitude_deg, pressure_millibars=standard_pressure_millibars, ozone_atm_cm=0.35, nitrogen_atm_cm=0.0002, precipitable_water_cm=5.0, turbidity_alpha=1.3, turbidity_beta=0.6):
    high = get_direct_normal_irradiance_by_band("high-frequency", altitude_deg, pressure_millibars,
                                           ozone_atm_cm, nitrogen_atm_cm, precipitable_water_cm, turbidity_alpha, turbidity_beta)
    low = get_direct_normal_irradiance_by_band("low-frequency", altitude_deg, pressure_millibars,
                                          ozone_atm_cm, nitrogen_atm_cm, precipitable_water_cm, turbidity_alpha, turbidity_beta)
    return high + low


def get_direct_normal_irradiance_by_band(band, altitude_deg, pressure_millibars=standard_pressure_millibars, ozone_atm_cm=0.35, nitrogen_atm_cm=0.0002, precipitable_water_cm=5.0, turbidity_alpha=1.3, turbidity_beta=0.6):
    ma = get_optical_mass_aerosol(altitude_deg)
    mo = get_optical_mass_ozone(altitude_deg)
    mR = get_optical_mass_rayleigh(altitude_deg, pressure_millibars)
    mRprime = mR * pressure_millibars / standard_pressure_millibars
    mw = get_optical_mass_water(altitude_deg)

    effective_wavelength = get_effective_aerosol_wavelength(
        band, ma, turbidity_alpha, turbidity_beta)
    tau_a = get_aerosol_optical_depth(
        turbidity_beta, effective_wavelength, turbidity_alpha)

    TR = get_rayleigh_transmittance(band, mRprime)
    Tg = get_gas_transmittance(band, mRprime)
    To = get_ozone_transmittance(band, mo, ozone_atm_cm)
    # is water_optical_mass really used for nitrogen calc?
    Tn = get_nitrogen_transmittance(band, mw, nitrogen_atm_cm)
    Tw = get_water_vapor_transmittance(band, mw, precipitable_water_cm)
    Ta = get_aerosol_transmittance(band, ma, tau_a)
    return E0n[band] * TR * Tg * To * Tn * Tw * Ta


def get_effective_aerosol_wavelength(band, ma, turbidity_alpha, turbidity_beta):
    # This function has an error somewhere. It returns negative values sometimes, but wavelength should always be positive.
    ua = math.log(1 + ma * turbidity_beta)
    if band == "high-frequency":
        a1 = turbidity_alpha  # just renaming to keep equations short
        d0 = 0.57664 - 0.024743 * a1
        d1 = (0.093942 - 0.2269 * a1 + 0.12848 * a1 ** 2) / (1 + 0.6418 * a1)
        d2 = (-0.093819 + 0.36668 * a1 - 0.12775 * a1 ** 2) / \
            (1 - 0.11651 * a1)
        d3 = a1 * (0.15232 - 0.087214 * a1 + 0.012664 * a1 ** 2) / \
            (1 - 0.90454 * a1 + 0.26167 * a1 ** 2)
        return (d0 + d1 * ua + d2 * ua ** 2) / (1 + d3 * ua ** 2)
    else:
        a2 = turbidity_alpha
        e0 = (1.183 - 0.022989 * a2 + 0.020829 * a2 ** 2) / (1 + 0.11133 * a2)
        e1 = (-0.50003 - 0.18329 * a2 + 0.23835 * a2 ** 2) / (1 + 1.6756 * a2)
        e2 = (-0.50001 + 1.1414 * a2 + 0.0083589 * a2 ** 2) / (1 + 11.168 * a2)
        e3 = (-0.70003 - 0.73587 * a2 + 0.51509 * a2 ** 2) / (1 + 4.7665 * a2)
        return (e0 + e1 * ua + e2 * ua ** 2) / (1 + e3 * ua)


def get_gas_transmittance(band, mRprime):
    if band == "high-frequency":
        return (1 + 0.95885 * mRprime + 0.012871 * mRprime ** 2) / (1 + 0.96321 * mRprime + 0.015455 * mRprime ** 2)
    else:
        return (1 + 0.27284 * mRprime - 0.00063699 * mRprime ** 2) / (1 + 0.30306 * mRprime)

def get_global_broadband_irradiance(altitude_deg, pressure_millibars=standard_pressure_millibars, ozone_atm_cm=0.35, nitrogen_atm_cm=0.0002, precipitable_water_cm=5.0, turbidity_alpha=1.3, turbidity_beta=0.6):
    Eb_high = get_beam_irradiance_by_band("high-frequency", altitude_deg, pressure_millibars, ozone_atm_cm, nitrogen_atm_cm, precipitable_water_cm, turbidity_alpha, turbidity_beta)
    Eb_low  = get_beam_irradiance_by_band("low-frequency",  altitude_deg, pressure_millibars, ozone_atm_cm, nitrogen_atm_cm, precipitable_water_cm, turbidity_alpha, turbidity_beta)

    Edp_high = get_diffuse_irradiance_by_band("high-frequency", air_mass, turbidity_alpha, turbidity_beta)
    Edp_low  = get_diffuse_irradiance_by_band("low-frequency",  air_mass, turbidity_alpha, turbidity_beta)

    Edd_high = get_backscattered_diffuse_irradiance_by_band("high-frequency", Eb_high, Edp_high, turbidity_alpha, turbidity_beta)
    Edd_low  = get_backscattered_diffuse_irradiance_by_band("low-frequency", Eb_low, Edp_low, turbidity_alpha, turbidity_beta)

    return Eb_high + Eb_low + Edp_high + Edp_low + Edd_high + Edd_low

def get_ground_albedo(band):
    # This could probably be improved with [Gueymard, 1993: Mathematically integrable parameterization of clear-sky beam and global irradiances and its use in daily irradiation applications]
    # http://www.sciencedirect.com/science/article/pii/0038092X9390059W]
    return 0.150  # mean ground albedo from [Gueymard, 2008], Table 1


def get_nitrogen_transmittance(band, mw, nitrogen_atm_cm):
    if band == "high-frequency":
        g1 = (0.17499 + 41.654 * un - 2146.4 * un ** 2) / \
            (1 + 22295.0 * un ** 2)
        g2 = un * (-1.2134 + 59.324 * un) / (1 + 8847.8 * un ** 2)
        g3 = (0.17499 + 61.658 * un + 9196.4 * un ** 2) / \
            (1 + 74109.0 * un ** 2)
        return min(1, (1 + g1 * mw + g2 * mw ** 2) / (1 + g3 * mw))
    else:
        return 1.0


# from Appendix B of [Gueymard, 2003]
def get_optical_mass_rayleigh(altitude_deg, pressure_millibars):
    Z = 90 - altitude_deg
    Z_rad = math.radians(Z)
    return (pressure_millibars / standard_pressure_millibars) / ((math.cos(Z_rad) + 0.48353 * Z_rad ** 0.095846) / (96.741 - Z_rad) ** 1.754)


def get_optical_mass_ozone(altitude_deg):  # from Appendix B of [Gueymard, 2003]
    Z = 90 - altitude_deg
    Z_rad = math.radians(Z)
    return 1 / ((math.cos(Z_rad) + 1.0651 * Z_rad ** 0.6379) / (101.8 - Z_rad) ** 2.2694)


def get_optical_mass_water(altitude_deg):  # from Appendix B of [Gueymard, 2003]
    Z = 90 - altitude_deg
    Z_rad = math.radians(Z)
    return 1 / ((math.cos(Z_rad) + 0.10648 * Z_rad ** 0.11423) / (93.781 - Z_rad) ** 1.9203)


def get_optical_mass_aerosol(altitude_deg):  # from Appendix B of [Gueymard, 2003]
    Z = 90 - altitude_deg
    Z_rad = math.radians(Z)
    return 1 / ((math.cos(Z_rad) + 0.16851 * Z_rad ** 0.18198) / (95.318 - Z_rad) ** 1.9542)


def get_ozone_transmittance(band, mo, uo):
    if band == "high-frequency":
        f1 = uo * (10.979 - 8.5421 * uo) / (1 + 2.0115 * uo + 40.189 * uo ** 2)
        f2 = uo * (-0.027589 - 0.005138 * uo) / \
            (1 - 2.4857 * uo + 13.942 * uo ** 2)
        f3 = uo * (10.995 - 5.5001 * uo) / (1 + 1.6784 * uo + 42.406 * uo ** 2)
        return (1 + f1 * mo + f2 * mo ** 2) / (1 + f3 * mo)
    else:
        return 1.0


def get_rayleigh_extinction_forward_scattering_fraction(band, mR):
    # returns BR
    if band == "high-frequency":
        return 0.5 * (0.89013 - 0.049558 * mR + 0.000045721 * mR ** 2)
    else:
        return 0.5


def get_rayleigh_transmittance(band, mRprime):
    if band == "high-frequency":
        return (1 + 1.8169 * mRprime + 0.033454 * mRprime ** 2) / (1 + 2.063 * mRprime + 0.31978 * mRprime ** 2)
    else:
        return (1 - 0.010394 * mRprime) / (1 - 0.00011042 * mRprime ** 2)


def get_sky_albedo(band, turbidity_alpha, turbidity_beta):
    if band == "high-frequency":
        a1 = turbidity_alpha  # just renaming to keep equations short
        b1 = turbidity_beta
        rhos = (0.13363 + 0.00077358 * a1 + b1 * (0.37567
                                                  + 0.22946 * a1) / (1 - 0.10832 * a1)) / (1 + b1 * (0.84057
                                                                                                     + 0.68683 * a1) / (1 - 0.08158 * a1))
    else:
        a2 = turbidity_alpha  # just renaming to keep equations short
        b2 = turbidity_beta
        rhos = (0.010191 + 0.00085547 * a2 + b2 * (0.14618
                                                   + 0.062758 * a2) / (1 - 0.19402 * a2)) / (1 + b2 * (0.58101
                                                                                                       + 0.17426 * a2) / (1 - 0.17586 * a2))
    return rhos


def get_water_vapor_transmittance(band, mw, w):
    if band == "high-frequency":
        h = get_water_vapor_transmittance_coefficients(band, w)
        return (1 + h[1] * mw) / (1 + h[2] * mw)
    else:
        c = get_water_vapor_transmittance_coefficients(band, w)
        return (1 + c[1] * mw + c[2] * mw ** 2) / (1 + c[3] * mw + c[4] * mw ** 2)


def get_water_vapor_transmittance_coefficients(band, w):
    if band == "high-frequency":
        h1 = w * (0.065445 + 0.00029901 * w) / (1 + 1.2728 * w)
        h2 = w * (0.065687 + 0.0013218 * w) / (1 + 1.2008 * w)
        return [float('NaN'), h1, h2]
    else:
        c1 = w * (19.566 - 1.6506 * w + 1.0672 * w ** 2) / \
            (1 + 5.4248 * w + 1.6005 * w ** 2)
        c2 = w * (0.50158 - 0.14732 * w + 0.047584 * w ** 2) / \
            (1 + 1.1811 * w + 1.0699 * w ** 2)
        c3 = w * (21.286 - 0.39232 * w + 1.2692 * w ** 2) / \
            (1 + 4.8318 * w + 1.412 * w ** 2)
        c4 = w * (0.70992 - 0.23155 * w + 0.096514 * w ** 2) / \
            (1 + 0.44907 * w + 0.75425 * w ** 2)
        return [float('NaN'), c1, c2, c3, c4]



import pysolar
from pysolar import radiation, solar
from pysolar import numeric as math
import datetime
import numpy as np
from nose.tools import raises, assert_equal


@raises(TypeError)
def test_fail_with_math():
    pysolar.use_math()
    lat = np.array([45., 40.])
    lon = np.array([3., 4.])
    time = datetime.datetime(2018, 5, 8, 12, 0, 0, tzinfo=datetime.timezone.utc)

    solar.get_altitude(lat, lon, time)


def test_scalar_with_math():
    pysolar.use_math()

    lat = 45.
    lon = 3.
    time = datetime.datetime(2018, 5, 8, 12, 0, 0, tzinfo=datetime.timezone.utc)

    print(solar.get_altitude(lat, lon, time))
    print(solar.get_azimuth(lat, lon, time))


def test_scalar_with_numpy():
    pysolar.use_numpy()

    lat = 50.63
    lon = 3.05
    time = datetime.datetime(2018, 5, 8, 12, 0, 0, tzinfo=datetime.timezone.utc)
    print(solar.get_altitude(lat, lon, time))
    print(solar.get_azimuth(lat, lon, time))


def test_with_fixed_time():
    """ get_altitude and get_azimuth, with scalar date """
    pysolar.use_numpy()

    lat = np.array([45., 40.])
    lon = np.array([3., 4.])

    time = datetime.datetime(2018, 5, 8, 12, 0, 0, tzinfo=datetime.timezone.utc)

    print(solar.get_altitude(lat, lon, time))
    print(solar.get_azimuth(lat, lon, time))
    print(solar.get_altitude_fast(lat, lon, time))
    print(solar.get_azimuth_fast(lat, lon, time))


def test_with_fixed_position():
    """ get_altitude and get_azimuth, with scalar position """
    pysolar.use_numpy()

    lat = 50.
    lon = 3.

    time = np.array(['2018-05-08T12:15:00',
                     '2018-05-08T15:00:00'], dtype='datetime64')

    print(solar.get_altitude_fast(lat, lon, time))
    print(solar.get_azimuth_fast(lat, lon, time))

def test_datetime_operations():

    d0 = datetime.datetime(2018,5,8,12,0,0)
    d1 = np.array(d0)

    assert_equal(math.tm_yday_math(d0),
                 math.tm_yday_numpy(d1))

    assert_equal(math.tm_hour_math(d0),
                 math.tm_hour_numpy(d1))

    assert_equal(math.tm_min_math(d0),
                 math.tm_min_numpy(d1))


def test_numpy():
    """ get_altitude and get_azimuth, with lat, lon and date arrays """
    pysolar.use_numpy()

    lat = np.array([45., 40.])
    lon = np.array([3., 4.])

    time = np.array(['2018-05-08T12:15:00',
                     '2018-05-08T15:00:00'], dtype='datetime64')

    print(solar.get_altitude_fast(lat, lon, time))
    print(solar.get_azimuth_fast(lat, lon, time))


def test_numpy_radiation():
    """
    get_radiation_direct with lat, lon, and date as arrays
    """
    pysolar.use_numpy()

    lat = np.array([45., 40., 40.])
    lon = np.array([3., 4., 3.])

    time = np.array([
        '2018-05-08T12:15:00',
        '2018-05-08T15:00:00',
        '2018-05-08T03:00:00',
    ], dtype='datetime64')

    altitude = solar.get_altitude_fast(lat, lon, time)
    rad_results = radiation.get_radiation_direct(time, altitude)
    assert rad_results[2] == 0
    print(rad_results)



#    Copyright Brandon Stafford
#
#    This file is part of Pysolar.
#
#    Pysolar is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Pysolar is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with Pysolar. If not, see <http://www.gnu.org/licenses/>.

"""Solar geometry functions

This module contains the most important functions for calculation of the position of the sun.

"""
from . import numeric as math
import datetime
from . import constants
from . import solartime as stime
from . import radiation
from .tzinfo_check import check_aware_dt


def solar_test():
    latitude_deg = 42.364908
    longitude_deg = -71.112828
    d = datetime.datetime.now(tz=datetime.timezone.utc)
    thirty_minutes = datetime.timedelta(hours = 0.5)
    for _ in range(48):
        timestamp = d.ctime()
        altitude_deg = get_altitude(latitude_deg, longitude_deg, d)
        azimuth_deg = get_azimuth(latitude_deg, longitude_deg, d)
        power = radiation.get_radiation_direct(d, altitude_deg)
        if (altitude_deg > 0):
            print(timestamp, "UTC", altitude_deg, azimuth_deg, power)
        d = d + thirty_minutes


def equation_of_time(day):
    "returns the number of minutes to add to mean solar time to get actual solar time."
    b = 2 * math.pi / 364.0 * (day - 81)
    return 9.87 * math.sin(2 * b) - 7.53 * math.cos(b) - 1.5 * math.sin(b)


def get_aberration_correction(sun_earth_distance):     # sun-earth distance is in astronomical units
    return -20.4898/(3600.0 * sun_earth_distance)


@check_aware_dt('when')
def get_topocentric_position(latitude_deg, longitude_deg, when, elevation = 0):
    '''Common calculations for altitude and azimuth'''
    # location-dependent calculations
    projected_radial_distance = get_projected_radial_distance(elevation, latitude_deg)
    projected_axial_distance = get_projected_axial_distance(elevation, latitude_deg)

    # time-dependent calculations
    jd = stime.get_julian_solar_day(when)
    jde = stime.get_julian_ephemeris_day(when)
    jce = stime.get_julian_ephemeris_century(jde)
    jme = stime.get_julian_ephemeris_millennium(jce)
    geocentric_latitude = get_geocentric_latitude(jme)
    geocentric_longitude = get_geocentric_longitude(jme)
    sun_earth_distance = get_sun_earth_distance(jme)
    aberration_correction = get_aberration_correction(sun_earth_distance)
    equatorial_horizontal_parallax = get_equatorial_horizontal_parallax(sun_earth_distance)
    nutation = get_nutation(jce)
    apparent_sidereal_time = get_apparent_sidereal_time(jd, jme, nutation)
    true_ecliptic_obliquity = get_true_ecliptic_obliquity(jme, nutation)

    # calculations dependent on location and time
    apparent_sun_longitude = get_apparent_sun_longitude(geocentric_longitude, nutation, aberration_correction)
    geocentric_sun_right_ascension = get_geocentric_sun_right_ascension(apparent_sun_longitude, true_ecliptic_obliquity, geocentric_latitude)
    geocentric_sun_declination = get_geocentric_sun_declination(apparent_sun_longitude, true_ecliptic_obliquity, geocentric_latitude)
    local_hour_angle = get_local_hour_angle(apparent_sidereal_time, longitude_deg, geocentric_sun_right_ascension)
    parallax_sun_right_ascension = get_parallax_sun_right_ascension(projected_radial_distance, equatorial_horizontal_parallax, local_hour_angle, geocentric_sun_declination)
    topocentric_local_hour_angle = get_topocentric_local_hour_angle(local_hour_angle, parallax_sun_right_ascension)
    topocentric_sun_declination = get_topocentric_sun_declination(geocentric_sun_declination, projected_axial_distance, equatorial_horizontal_parallax, parallax_sun_right_ascension, local_hour_angle)

    return topocentric_sun_declination, topocentric_local_hour_angle


@check_aware_dt('when')
def get_position(latitude_deg, longitude_deg, when, elevation=0,
                 temperature = constants.standard_temperature,
                 pressure = constants.standard_pressure):
    ''' Given location, time and atmospheric conditions
    temperature in Kelvin and pressure in Pascal
    
    returns (azimuth, altitude) of sun in degrees.

    Same as a combination of get_azimuth and get_altitude
    '''

    topocentric_sun_declination, topocentric_local_hour_angle = \
        get_topocentric_position(latitude_deg, longitude_deg, when, elevation)

    topocentric_elevation_angle = \
        get_topocentric_elevation_angle(latitude_deg, topocentric_sun_declination,
                                        topocentric_local_hour_angle)

    refraction_correction = get_refraction_correction(pressure, temperature,
                                                      topocentric_elevation_angle)

    altitude_deg = topocentric_elevation_angle + refraction_correction

    azimuth_deg = get_topocentric_azimuth_angle(topocentric_local_hour_angle,
                                      latitude_deg, topocentric_sun_declination)

    return azimuth_deg, altitude_deg


@check_aware_dt('when')
def get_altitude(latitude_deg, longitude_deg, when, elevation = 0,
                 temperature = constants.standard_temperature, pressure = constants.standard_pressure):
    '''See also the faster, but less accurate, get_altitude_fast()
    temperature in Kelvin and pressure in Pascal
    '''
    topocentric_sun_declination, topocentric_local_hour_angle = \
        get_topocentric_position(latitude_deg, longitude_deg, when)

    topocentric_elevation_angle = get_topocentric_elevation_angle(latitude_deg, topocentric_sun_declination, topocentric_local_hour_angle)
    refraction_correction = get_refraction_correction(pressure, temperature, topocentric_elevation_angle)
    return topocentric_elevation_angle + refraction_correction


@check_aware_dt('when')
def get_altitude_fast(latitude_deg, longitude_deg, when):
# expect 19 degrees for solar.get_altitude(42.364908,-71.112828,datetime.datetime(2007, 2, 18, 20, 13, 1, 130320))
    day = math.tm_yday(when)
    declination_rad = math.radians(get_declination(day))
    latitude_rad = math.radians(latitude_deg)
    hour_angle = get_hour_angle(when, longitude_deg)
    first_term = math.cos(latitude_rad) * math.cos(declination_rad) * math.cos(math.radians(hour_angle))
    second_term = math.sin(latitude_rad) * math.sin(declination_rad)
    return math.degrees(math.asin(first_term + second_term))


def get_apparent_sidereal_time(jd, jme, nutation):
    return get_mean_sidereal_time(jd) + nutation['longitude'] * math.cos(get_true_ecliptic_obliquity(jme, nutation))


def get_apparent_sun_longitude(geocentric_longitude, nutation, ab_correction):
    return geocentric_longitude + nutation['longitude'] + ab_correction


@check_aware_dt('when')
def get_azimuth(latitude_deg, longitude_deg, when, elevation = 0):

    topocentric_sun_declination, topocentric_local_hour_angle = \
        get_topocentric_position(latitude_deg, longitude_deg, when, elevation)

    azimuth = get_topocentric_azimuth_angle(topocentric_local_hour_angle,
            latitude_deg, topocentric_sun_declination)

    return azimuth


def get_azimuth_fast(latitude_deg, longitude_deg, when):
# expect 230 degrees for solar.get_azimuth(42.364908,-71.112828,datetime.datetime(2007, 2, 18, 20, 18, 0, 0))
    day = math.tm_yday(when)
    declination_rad = math.radians(get_declination(day))
    latitude_rad = math.radians(latitude_deg)
    hour_angle_rad = math.radians(get_hour_angle(when, longitude_deg))
    altitude_rad = math.radians(get_altitude_fast(latitude_deg, longitude_deg, when))

    azimuth_rad = math.asin(-math.cos(declination_rad) * math.sin(hour_angle_rad) / math.cos(altitude_rad))

    return math.where(math.cos(hour_angle_rad) * math.tan(latitude_rad) >= math.tan(declination_rad),
                      (180 - math.degrees(azimuth_rad)),
                      math.degrees(azimuth_rad) + 360 * (azimuth_rad < 0)
                     )

def get_coeff(jme, coeffs):
    "computes a polynomial with time-varying coefficients from the given constant" \
    " coefficients array and the current Julian millennium."
    result = 0.0
    x = 1.0
    for line in coeffs :
        c = 0.0
        for l in line :
            c += l[0] * math.cos(l[1] + l[2] * jme)
        #end for
        result += c * x
        x *= jme
    #end for
    return \
        result
#end get_coeff

def get_declination(day):
    '''The declination of the sun is the angle between
    Earth's equatorial plane and a line between the Earth and the sun.
    The declination of the sun varies between 23.45 degrees and -23.45 degrees,
    hitting zero on the equinoxes and peaking on the solstices.
    '''
    return constants.earth_axis_inclination * math.sin((2 * math.pi / 365.0) * (day - 81))

def get_equatorial_horizontal_parallax(sun_earth_distance):
    return 8.794 / (3600 / sun_earth_distance)

def get_flattened_latitude(latitude):
    latitude_rad = math.radians(latitude)
    return math.degrees(math.atan(0.99664719 * math.tan(latitude_rad)))

# Geocentric functions calculate angles relative to the center of the earth.

def get_geocentric_latitude(jme):
    return -1 * get_heliocentric_latitude(jme)

def get_geocentric_longitude(jme):
    return (get_heliocentric_longitude(jme) + 180) % 360

def get_geocentric_sun_declination(apparent_sun_longitude, true_ecliptic_obliquity, geocentric_latitude):
    apparent_sun_longitude_rad = math.radians(apparent_sun_longitude)
    true_ecliptic_obliquity_rad = math.radians(true_ecliptic_obliquity)
    geocentric_latitude_rad = math.radians(geocentric_latitude)

    a = math.sin(geocentric_latitude_rad) * math.cos(true_ecliptic_obliquity_rad)
    b = math.cos(geocentric_latitude_rad) * math.sin(true_ecliptic_obliquity_rad) * math.sin(apparent_sun_longitude_rad)
    delta = math.asin(a + b)
    return math.degrees(delta)

def get_geocentric_sun_right_ascension(apparent_sun_longitude, true_ecliptic_obliquity, geocentric_latitude):
    apparent_sun_longitude_rad = math.radians(apparent_sun_longitude)
    true_ecliptic_obliquity_rad = math.radians(true_ecliptic_obliquity)
    geocentric_latitude_rad = math.radians(geocentric_latitude)

    a = math.sin(apparent_sun_longitude_rad) * math.cos(true_ecliptic_obliquity_rad)
    b = math.tan(geocentric_latitude_rad) * math.sin(true_ecliptic_obliquity_rad)
    c = math.cos(apparent_sun_longitude_rad)
    alpha = math.atan2((a - b),  c)
    return math.degrees(alpha) % 360

# Heliocentric functions calculate angles relative to the center of the sun.

def get_heliocentric_latitude(jme):
    return math.degrees(get_coeff(jme, constants.heliocentric_latitude_coeffs) / 1e8)

def get_heliocentric_longitude(jme):
    return math.degrees(get_coeff(jme, constants.heliocentric_longitude_coeffs) / 1e8) % 360

@check_aware_dt('when')
def get_hour_angle(when, longitude_deg):
    solar_time = get_solar_time(longitude_deg, when)
    return 15.0 * (solar_time - 12.0)

def get_incidence_angle(topocentric_zenith_angle, slope, slope_orientation, topocentric_azimuth_angle):
    tza_rad = math.radians(topocentric_zenith_angle)
    slope_rad = math.radians(slope)
    so_rad = math.radians(slope_orientation)
    taa_rad = math.radians(topocentric_azimuth_angle)
    return math.degrees(math.acos(math.cos(tza_rad) * math.cos(slope_rad) + math.sin(slope_rad) * math.sin(tza_rad) * math.cos(taa_rad - math.pi - so_rad)))

def get_local_hour_angle(apparent_sidereal_time, longitude, geocentric_sun_right_ascension):
    return (apparent_sidereal_time + longitude - geocentric_sun_right_ascension) % 360

def get_mean_sidereal_time(jd):
    # This function doesn't agree with Andreas and Reda as well as it should. Works to ~5 sig figs in current unit test
    jc = stime.get_julian_century(jd)
    sidereal_time =  280.46061837 + (360.98564736629 * (jd - 2451545.0)) + 0.000387933 * jc * jc * (1 - jc / 38710000)
    return sidereal_time % 360

def get_nutation(jce):
    abcd = constants.nutation_coefficients
    nutation_long = []
    nutation_oblique = []
    p = constants.get_aberration_coeffs()
    x = list \
      (
        p[k](jce)
        for k in
            ( # order is important
                'MeanElongationOfMoon',
                'MeanAnomalyOfSun',
                'MeanAnomalyOfMoon',
                'ArgumentOfLatitudeOfMoon',
                'LongitudeOfAscendingNode',
            )
      )
    y = constants.aberration_sin_terms
    for i in range(len(abcd)):
        sigmaxy = 0.0
        for j in range(len(x)):
            sigmaxy += x[j] * y[i][j]
        #end for
        nutation_long.append((abcd[i][0] + (abcd[i][1] * jce)) * math.sin(math.radians(sigmaxy)))
        nutation_oblique.append((abcd[i][2] + (abcd[i][3] * jce)) * math.cos(math.radians(sigmaxy)))

    # 36000000 scales from 0.0001 arcseconds to degrees
    nutation = {'longitude' : sum(nutation_long)/36000000.0, 'obliquity' : sum(nutation_oblique)/36000000.0}

    return nutation
#end get_nutation

def get_parallax_sun_right_ascension(projected_radial_distance, equatorial_horizontal_parallax, local_hour_angle, geocentric_sun_declination):
    prd = projected_radial_distance
    ehp_rad = math.radians(equatorial_horizontal_parallax)
    lha_rad = math.radians(local_hour_angle)
    gsd_rad = math.radians(geocentric_sun_declination)
    a = -1 * prd * math.sin(ehp_rad) * math.sin(lha_rad)
    b =  math.cos(gsd_rad) - prd * math.sin(ehp_rad) * math.cos(lha_rad)
    parallax = math.atan2(a, b)
    return math.degrees(parallax)

def get_projected_radial_distance(elevation, latitude):
    flattened_latitude_rad = math.radians(get_flattened_latitude(latitude))
    latitude_rad = math.radians(latitude)
    return math.cos(flattened_latitude_rad) + (elevation * math.cos(latitude_rad) / constants.earth_radius)

def get_projected_axial_distance(elevation, latitude):
    flattened_latitude_rad = math.radians(get_flattened_latitude(latitude))
    latitude_rad = math.radians(latitude)
    return 0.99664719 * math.sin(flattened_latitude_rad) + (elevation * math.sin(latitude_rad) / constants.earth_radius)

def get_sun_earth_distance(jme):
    return get_coeff(jme, constants.sun_earth_distance_coeffs) / 1e8

def get_refraction_correction(pressure, temperature, topocentric_elevation_angle):
    #function and default values according to original NREL SPA C code
    #http://www.nrel.gov/midc/spa/

    sun_radius = 0.26667
    atmos_refract = 0.5667
    tea = topocentric_elevation_angle

    # Approximation only valid if sun is not well below horizon
    # This approximation could be improved; see history at https://github.com/pingswept/pysolar/pull/23
    # Better method could come from Auer and Standish [2000]:
    # http://iopscience.iop.org/1538-3881/119/5/2472/pdf/1538-3881_119_5_2472.pdf

    a = pressure * 2.830 * 1.02
    b = 1010.0 * temperature * 60.0 * math.tan(math.radians(tea + (10.3/(tea + 5.11))))

    del_e = math.where(tea >= -1.0*(sun_radius + atmos_refract),
                       a / b, 0.)

    return del_e

@check_aware_dt('when')
def get_solar_time(longitude_deg, when):
    "returns solar time in hours for the specified longitude and time," \
    " accurate only to the nearest minute."
    return \
        (
            (math.tm_hour(when) * 60 + math.tm_min(when) + 4 * longitude_deg + equation_of_time(math.tm_yday(when)))
        /
            60
        )

# Topocentric functions calculate angles relative to a location on the surface of the earth.

def get_topocentric_azimuth_angle(topocentric_local_hour_angle, latitude, topocentric_sun_declination):
    """West is negative, East is positive, Masters p. 395"""
    tlha_rad = math.radians(topocentric_local_hour_angle)
    latitude_rad = math.radians(latitude)
    tsd_rad = math.radians(topocentric_sun_declination)
    a = math.sin(tlha_rad)
    b = math.cos(tlha_rad) * math.sin(latitude_rad) - math.tan(tsd_rad) * math.cos(latitude_rad)
    return (180.0 + math.degrees(math.atan2(a, b))) % 360

def get_topocentric_elevation_angle(latitude, topocentric_sun_declination, topocentric_local_hour_angle):
    latitude_rad = math.radians(latitude)
    tsd_rad = math.radians(topocentric_sun_declination)
    tlha_rad = math.radians(topocentric_local_hour_angle)
    return math.degrees(math.asin((math.sin(latitude_rad) * math.sin(tsd_rad)) + math.cos(latitude_rad) * math.cos(tsd_rad) * math.cos(tlha_rad)))

def get_topocentric_local_hour_angle(local_hour_angle, parallax_sun_right_ascension):
    return local_hour_angle - parallax_sun_right_ascension

def get_topocentric_sun_declination(geocentric_sun_declination, projected_axial_distance, equatorial_horizontal_parallax, parallax_sun_right_ascension, local_hour_angle):
    gsd_rad = math.radians(geocentric_sun_declination)
    pad = projected_axial_distance
    ehp_rad = math.radians(equatorial_horizontal_parallax)
    psra_rad = math.radians(parallax_sun_right_ascension)
    lha_rad = math.radians(local_hour_angle)
    a = (math.sin(gsd_rad) - pad * math.sin(ehp_rad)) * math.cos(psra_rad)
    b = math.cos(gsd_rad) - (pad * math.sin(ehp_rad) * math.cos(lha_rad))
    return math.degrees(math.atan2(a, b))

def get_topocentric_sun_right_ascension(projected_radial_distance, equatorial_horizontal_parallax, local_hour_angle,
        apparent_sun_longitude, true_ecliptic_obliquity, geocentric_latitude):
    gsd = get_geocentric_sun_declination(apparent_sun_longitude, true_ecliptic_obliquity, geocentric_latitude)
    psra = get_parallax_sun_right_ascension(projected_radial_distance, equatorial_horizontal_parallax, local_hour_angle, gsd)
    gsra = get_geocentric_sun_right_ascension(apparent_sun_longitude, true_ecliptic_obliquity, geocentric_latitude)
    return psra + gsra

def get_topocentric_zenith_angle(latitude, topocentric_sun_declination, topocentric_local_hour_angle, pressure, temperature):
    tea = get_topocentric_elevation_angle(latitude, topocentric_sun_declination, topocentric_local_hour_angle)
    return 90 - tea - get_refraction_correction(pressure, temperature, tea)

def get_true_ecliptic_obliquity(jme, nutation):
    u = jme/10.0
    mean_obliquity = 84381.448 - (4680.93 * u) - (1.55 * u ** 2) + (1999.25 * u ** 3) \
    - (51.38 * u ** 4) -(249.67 * u ** 5) - (39.05 * u ** 6) + (7.12 * u ** 7) \
    + (27.87 * u ** 8) + (5.79 * u ** 9) + (2.45 * u ** 10)
    return (mean_obliquity / 3600.0) + nutation['obliquity']
